# Study the cases where the covering degree is NOT 1.

Read("classical_group.g");
Read("amalgam.g");

bad_cases := [
    # Sp_6(2) = W(5,2) \cong  O_7(2) = Q(6,2) 
    rec(
        deg  := infinity,
        name := "Sp_6(2) = W(5,2) isomorphic to O_7(2) = Q(6,2)",
        unip := UnipotentRootGroups(GF(2), 3, 0, 1, -1),
    ),
    # Sp_6(3) = W(5,3)
    rec(
        deg  := infinity,
        name := "Sp_6(3) = W(5,3)",
        unip := UnipotentRootGroups(GF(3), 3, 0, 1, -1),
    ),
    # O_7(2) = Q(6,2) \cong Sp_6(2) = W(5,2)
    rec(
        deg  := infinity,
        name := "O_7(2) = Q(6,2) isomorphic to Sp_6(2) = W(5,2)",
        unip := UnipotentRootGroups(GF(2), 3, 1, 1, 1),
    ),
    # O_7(3) = Q(6,3)
    rec(
        deg  := 81,
        name := "O_7(3) = Q(6,3)",
        unip := UnipotentRootGroups(GF(3), 3, 1, 1, 1),
    ),
    # O_7(4) = Q(6,4)
    rec(
        deg  := 4,
        name := "O_7(4) = Q(6,4)",
        unip := UnipotentRootGroups(GF(4), 3, 1, 1, 1),
    ),
    # O_8(2)  = Q^-(7,2)
    rec(
        deg  := infinity,
        name := "O_8(2)  = Q^-(7,2)",
        unip := UnipotentRootGroups(GF(2), 3, 2, 1, 1),
    ),
    # U_6(2) = H(5, 2^2)
    rec(
        deg  := infinity,
        name := "U_6(2) = H(5, 2^2)",
        unip := UnipotentRootGroups(GF(2^2), 3, 0, 2, 1),
    ),
];;


for bc in bad_cases do
    Print("\n---- Looking at '", bc.name, "'\n");
    Print("  ",bc.deg," is the expected covering degree\n");

    # Rewrite those root groups and the group U they generate as pcp groups,
    # exploiting their polycyclic structure.
    root_groups := RootGroups_to_PcpGroups( bc.unip[2] );
    U := Parent(root_groups[1]);

    # construct the amalgamated product Utild
    ama := AmalgamOfRootGroups(root_groups);
    Utilde := ama[1];

    if bc.deg = infinity then
        # construct the second derived subgroup
        g := DerivedSubgroup(DerivedSubgroup(Utilde));
    
        # Utilde is certainly infinite if g is infinite, which is in turn
        # implied by 0 being among its abelian invariants
        ai := AbelianInvariants(g);
        Print("    abelian invariants of Utilde'' = ", ai, "\n");
        if 0 in ai then
            Print("   covering degree infinity VERIFIED\n");
        else
            Print("   ERROR\n");
        fi;
    else
        # Verify the expected covering degree of Utilde over U using coset enumeration.
        # A (small) finite covering degree is expected.
        U12 := Subgroup(U, Concatenation(List(ama[2]{[1,2]}, hom->GeneratorsOfGroup(PreImagesRange(hom)))));
        U12tilde := Subgroup(Utilde, Concatenation(List(ama[2]{[1,2]}, hom->GeneratorsOfGroup(ImagesSource(hom)))));
        idx := Index(Utilde, U12tilde);
        SetSize(Utilde, Size(U12) * idx);
        deg := Size(Utilde) / Size(U);
        
        if bc.deg <> deg then
            Print("  ERROR: found covering degree ", deg, " instead of ", bc.deg, "\n");
            continue;
        fi;

        # Compute maximal nilpotent quotient Unq of Utilde and verify they are isomorphic
        if not IsEmpty(OptionsStack) then
            ResetOptionsStack(); # work around strange left overs on stack
        fi;
        epi := NqEpimorphismNilpotentQuotient(Utilde);
        Unq := Range(epi);
        
        if Size(Utilde) <> Size(Unq) then
            Print("  ERROR: covering degree matches but Utilde is not nilpotent\n");
            continue;
        fi;
        
        # construct the covering map from Unq onto U
        gens := [];
        imgs := [];
        for hom in ama[2] do
            for g in GeneratorsOfGroup(Image(hom)) do
                Add(gens, ImageElm(epi, g)); # map this from Utilde to Unq
                Add(imgs, PreImagesRepresentative(hom, g));
            od;
        od;
        phi := GroupHomomorphismByImages(Unq, U, gens, imgs);

        # Now we can study properties of Unq relative to U
        Print("  Some properties of the covering map phi : Utilde -> U\n");
        Print("    phi is surjective: ", IsSurjective(phi), "\n");

        Print("    |Unq| = ", Size(Unq), ";\n      class ", NilpotencyClassOfGroup(Unq),
                ";\n      generators ", AsList(Pcp(Unq)), "\n");

        K := Kernel(phi);
        Print("    K := ker(phi) = ", StructureDescription(K), ";\n     small groups id ", IdGroup(K),
                ";\n      generators ", AsList(Pcp(K)), "\n");

        z:=Center(Unq);
        Print("    Center(Unq) = ", StructureDescription(z), ";\n      small groups id ", IdGroup(z),
                ";\n      generators ", AsList(Pcp(z)), "\n");

        cK := Centralizer(Unq, K);
        Print("    |C_Unq(K)| = ", Size(cK), ";\n      class ", NilpotencyClassOfGroup(cK),
                ";\n      generators ", AsList(Pcp(cK)), "\n");

        nK := Normalizer(Unq, K);
        Print("    |N_Unq(K)| = ", Size(nK), ";\n      class ", NilpotencyClassOfGroup(nK),
                ";\n      generators ", AsList(Pcp(nK)), "\n");

        # check for a splitting
        ccr := ComplementClassesRepresentatives(Unq, K);
        Print("    extension splits: ", not IsEmpty(ccr), "\n");
        
        if IsElementaryAbelian(K) then
            p := PrimePGroup(K);
            n := RankPGroup(K);
            Print("    Unq is extension of U over module F_",p,"^",n,"\n");

            cr:=CRRecordBySubgroup(Unq, K);;
            cc:=TwoCohomologyCR(cr);;
            Assert(0, ForAll(cc.factor.rels, x -> x = p));
            Print("    H^2(U, K) = F_",p,"^",Length(cc.factor.rels),"\n");
        fi;
        
    fi;
od;
